/* --------------------------------------------------------------------------
 * module_cdrom.c
 * code for module cdrom
 *
 * Copyright 2002 Matthias Grimm (joker@cymes.de).
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 * -------------------------------------------------------------------------*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <linux/cdrom.h>

#include <pbb.h>

#include "pbbinput.h"
#include "gettext_macros.h"
#include "input_manager.h"
#include "module_cdrom.h"
#include "support.h"

struct moddata_cdrom {
	char *cdromdev;		/* pathname of the cdrom device */
	unsigned short keyejectcd;
	unsigned short modejectcd;
	struct timeval tv;      /* time of key press */
	int keydelaytime;      /* delay in ms */
	int cdejected;
} modbase_cdrom;

int
cdrom_init (struct tagitem *taglist)
{
	struct moddata_cdrom *base = &modbase_cdrom;
	static char devbuffer_cdrom[STDBUFFERLEN];
	int rc;

	sprintf(devbuffer_cdrom, DEFAULT_CDROM);

	base->cdromdev		= devbuffer_cdrom;
	base->keyejectcd		= KEY_EJECTCD;
	base->modejectcd		= MOD_NONE;
	base->keydelaytime		= 0;

	if ((rc = cdrom_handle_tags (MODE_CONFIG, taglist))) /* interprete configfile */
		return rc;                 /* critcal error occured, abort program */

	register_function (KBDQUEUE, cdrom_keyboard);
	register_function (QUERYQUEUE, cdrom_query);
	register_function (CONFIGQUEUE, cdrom_configure);
	return 0;
}

int
cdrom_exit ()
{
	return 0;
}

void
cdrom_keyboard (struct tagitem *taglist)
{
	struct moddata_cdrom *base = &modbase_cdrom;
	int code, value, mod;

	code = (int) tagfind (taglist, TAG_KEYCODE, 0);
	value = (int) tagfind (taglist, TAG_KEYREPEAT, 0);
	mod = (int) tagfind (taglist, TAG_MODIFIER, 0);

	if (value) {
		if (value == 1) base->cdejected = 0;
		if ((code == base->keyejectcd) && (mod == base->modejectcd))
			if ((base->keydelaytime == 0) || (keydelayms(&base->tv, value, base->keydelaytime)))
				if (base->cdejected == 0) {
					base->cdejected = 1;
					cdrom_eject();
				}
	}
}

void
cdrom_query (struct tagitem *taglist)
{
	cdrom_handle_tags(MODE_QUERY, taglist);
}

void
cdrom_configure (struct tagitem *taglist)
{
	cdrom_handle_tags(MODE_CONFIG, taglist);
}

int
cdrom_handle_tags (int cfgure, struct tagitem *taglist)
{
	struct moddata_cdrom *base = &modbase_cdrom;
	int err, rc = 0;

	while (taglist->tag != TAG_END) {
		switch (taglist->tag) {
		case TAG_EJECTCD:
			if (cfgure)	cdrom_eject();
			else		tagerror (taglist, E_NOREAD);
			break;
		case TAG_CDROMDEVICE:
			if (cfgure) {
				if ((err = copy_path ((char *) taglist->data, base->cdromdev, TYPE_BLKDEV, CPFLG_MAYBEMISSED)))
					rc = tagerror (taglist, err);
			} else
				taglist->data = (long) base->cdromdev;
			break;
		case TAG_EJECTCDKEY:
			if (cfgure)	base->keyejectcd = taglist->data;
			else		taglist->data = base->keyejectcd;
			break;
		case TAG_EJECTCDMOD:
			if (cfgure)	base->modejectcd = taglist->data;
			else		taglist->data = base->modejectcd;
			break;
		case TAG_EJECTCDKEYDELAY:
			if (cfgure)	base->keydelaytime = taglist->data;
			else		taglist->data = base->keydelaytime;
			break;
		}
		taglist++;
	}
	return rc;
}

void
cdrom_eject()
{
	struct moddata_cdrom *base = &modbase_cdrom;
	int fd, err;

	err = call_script ("/bin/umount %.30s", base->cdromdev);
	if (!err || err == E_CLDEXIT)
		if ((fd = open(base->cdromdev, O_RDONLY | O_NDELAY)) >= 0) {
			ioctl (fd, CDROMEJECT);
			close (fd);
		}
}
