/* --------------------------------------------------------------------------
 * module_display.c
 * code for module display
 *
 * Copyright 2002 Matthias Grimm (joker@cymes.de).
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 * -------------------------------------------------------------------------*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include "pbbinput.h"

#include <pbb.h>

#include "input_manager.h"
#include "module_display.h"
#include "support.h"

struct moddata_display {
	char *fbdev;	/* pathname of the framebuffer device */
	int ctrl_fb;	 /* 1, if framebuffer should be blanked too */
	int dimmed;                /* =1, when backlight is dimmed, =2, when it is switched off */
	int dimrequest;
	unsigned short keybrightup;	/* keycodes and modifiers for controls */
	unsigned short modbrightup;
	unsigned short keybrightdn;
	unsigned short modbrightdn;
	int backlightmax;           /* maximum brightness level of backlight */
	int fadingspeed;          /* interval in 10th ms for fading brightness */
	int dimfullydark;
	int currentlevel;       /* current brightness level */
	int savedlevel;         /* saved brightness level during dim or blank */
	int targetlevel;      /* brightness level to set up at the end */
} modbase_display;

int
display_init (struct tagitem *taglist)
{
	struct moddata_display *base = &modbase_display;
	struct tagitem args[] = {{ TAG_BACKLIGHTMAX, -1 },
					      { TAG_BACKLIGHTLEVEL, -1 },
					      { TAG_END, 0 }};
	int rc;

	static char devbuffer_fb[STDBUFFERLEN];
	sprintf(devbuffer_fb, DEFAULT_FB_DEVICE);

	process_queue (QUERYQUEUE, args);
	base->fbdev			= devbuffer_fb;
	base->ctrl_fb		= 0;            /* disabled by default  */
	base->currentlevel	= args[1].data;
	base->savedlevel	= args[1].data;
	base->targetlevel	= -1;
	base->dimmed		= 0;
	base->dimrequest	= 0;
	base->keybrightup	= KEY_BRIGHTNESSUP;
	base->modbrightup	= MOD_NONE;
	base->keybrightdn	= KEY_BRIGHTNESSDOWN;
	base->modbrightdn	= MOD_NONE;
	base->backlightmax	= args[0].data;
	base->fadingspeed	= 0;     /* fading disabled by default */
	base->dimfullydark  = 0;

	if ((rc = display_handle_tags (MODE_CONFIG, taglist))) /* interprete configfile */
		return rc;         /* critcal error occured, abort program */

	register_function (KBDQUEUE, display_keyboard);
	register_function (QUERYQUEUE, display_query);
	register_function (CONFIGQUEUE, display_configure);
	register_function (T10QUEUE, display_timer);
	return 0;
}

int
display_exit ()
{
	return 0;
}

void
display_keyboard (struct tagitem *taglist)
{
	struct moddata_display *base = &modbase_display;
	int code, value, mod, level;

	code = (int) tagfind (taglist, TAG_KEYCODE, 0);
	value = (int) tagfind (taglist, TAG_KEYREPEAT, 0);
	mod = (int) tagfind (taglist, TAG_MODIFIER, 0);

	if (value) {
		if ((code == base->keybrightup) && ((mod & ~MOD_SHIFT) == base->modbrightup))
			level = display_change_brightness (mod & MOD_SHIFT ? OP_BRIGHTNESS_MAX : OP_BRIGHTNESS_UP);
		else if ((code == base->keybrightdn) && ((mod & ~MOD_SHIFT) == base->modbrightdn))
			level = display_change_brightness (mod & MOD_SHIFT ? OP_BRIGHTNESS_MIN : OP_BRIGHTNESS_DOWN);
		else return;
		singletag_to_clients(CHANGEVALUE, TAG_BRIGHTNESS, level);
	}
}

void
display_timer (struct tagitem *taglist)
{
	struct moddata_display *base = &modbase_display;
	static int count = 0;
	int level;

	if (base->targetlevel >= 0) {   /* level should be changed */
		count++;
		if (count >= base->fadingspeed) {
			count = 0;
			if (base->fadingspeed) {
				level = base->currentlevel;
				if (level > base->targetlevel)
					level--;   /* target is darker as current */
				else
					level++;   /* target is brighter as current */
			}else
				level = base->targetlevel;

			process_queue_single (CONFIGQUEUE, TAG_BACKLIGHTLEVEL, level);
			if (((level == 0) && (base->currentlevel != 0)) ||
			    ((level != 0) && (base->currentlevel == 0)))
				display_framebuffer (level ? 1: 0);
			base->currentlevel = level;  /* target level becomes current level */
			base->dimmed = base->dimrequest;

			if (level == base->targetlevel)
				base->targetlevel = -1;
		}
  }
}

void
display_query (struct tagitem *taglist)
{
	display_handle_tags (MODE_QUERY, taglist);
}

void
display_configure (struct tagitem *taglist)
{
	display_handle_tags (MODE_CONFIG, taglist);
}

int
display_handle_tags (int cfgure, struct tagitem *taglist)
{
	struct moddata_display *base = &modbase_display;
	int rc = 0, err;

	while (taglist->tag != TAG_END) {
		switch (taglist->tag) {
		case TAG_BRIGHTNESSOP:  /* private tag */
			if (cfgure)	display_change_brightness (taglist->data);
			break;
		case TAG_BRIGHTNESS:
			if (cfgure) {
				display_set_brightness (taglist->data);
				singletag_to_clients(CHANGEVALUE, TAG_BRIGHTNESS, taglist->data);
			} else
				taglist->data = process_queue_single (QUERYQUEUE, TAG_BACKLIGHTLEVEL, -1);
			break;
		case TAG_BRIGHTNESSUPKEY:
			if (cfgure)	base->keybrightup = taglist->data;
			else		taglist->data = base->keybrightup;
			break;
		case TAG_BRIGHTNESSUPMOD:
			if (cfgure)	base->modbrightup = taglist->data;
			else		taglist->data = base->modbrightup;
			break;
		case TAG_BRIGHTNESSDOWNKEY:
			if (cfgure)	base->keybrightdn = taglist->data;
			else		taglist->data = base->keybrightdn;
			break;
		case TAG_BRIGHTNESSDOWNMOD:
			if (cfgure)	base->modbrightdn = taglist->data;
			else		taglist->data = base->modbrightdn;
			break;
		case TAG_FRAMEBUFFERDEVICE:
			if (cfgure) {
				if ((err = copy_path ((char *) taglist->data, base->fbdev, TYPE_CHARDEV, CPFLG_NONE)))
					rc = tagerror (taglist, err);
			} else
				taglist->data = (long) base->fbdev;
			break;
		case TAG_BLANKFRAMEBUFFER:
			if (cfgure)	base->ctrl_fb = taglist->data;
			else		taglist->data = base->ctrl_fb;
			break;
		case TAG_BRIGHTNESSMAX:
			if (cfgure)	tagerror (taglist, E_NOWRITE);
			else		taglist->data = base->backlightmax;
			break;
		case TAG_FADINGSPEED:
			if (cfgure)	base->fadingspeed = taglist->data;
			else		taglist->data = base->fadingspeed;
			break;
		case TAG_DIMFULLYDARK:
			if (cfgure)	base->dimfullydark = taglist->data;
			else		taglist->data = base->dimfullydark;
			break;
		}
		taglist++;
	}
	return rc;
}

void
display_set_brightness(int level)
{
	struct moddata_display *base = &modbase_display;

	if (level > BACKLIGHT_MAX)
		level = BACKLIGHT_MAX;
	if (level < BACKLIGHT_OFF)
		level = BACKLIGHT_OFF;
	base->dimrequest = 0;
	base->targetlevel = level;  /* brightness level set in timer function */
}

int
display_change_brightness(int op)
{
	struct moddata_display *base = &modbase_display;
	int level;

	if (((op == OP_DIM_RECOVER) && (base->dimmed == 0)) ||
	     ((op == OP_DIM_LIGHT) && (base->dimmed >= 1)) ||
	     ((op == OP_DIM_OFF) && (base->dimmed == 2)))
		return base->currentlevel;

	if ((base->currentlevel = process_queue_single (QUERYQUEUE, TAG_BACKLIGHTLEVEL, -1)) >= 0) {
		if ((base->currentlevel > 1) && (base->dimmed)) /* someone else increased brightness? */
			base->dimmed = 0;             /* so we can't stay dimmed. */
		level = (base->dimmed) ? base->savedlevel : base->currentlevel;

		if (base->dimfullydark && op == OP_DIM_LIGHT)
			op = OP_DIM_OFF;

		switch (op) {
		case OP_BRIGHTNESS_UP:
			if (level < BACKLIGHT_MAX)
				level++;
			base->dimrequest = 0;
			break;
		case OP_BRIGHTNESS_DOWN:
			if (level > BACKLIGHT_OFF)
				level--;
			base->dimrequest = 0;
			break;
		case OP_BRIGHTNESS_MIN:
			level = BACKLIGHT_MIN;
			base->dimrequest = 0;
			break;
		case OP_BRIGHTNESS_MAX:
			level = BACKLIGHT_MAX;
			base->dimrequest = 0;
			break;
		case OP_DIM_LIGHT:
			base->savedlevel = base->currentlevel;
			if (level > BACKLIGHT_OFF) /* if level = 0 then let it as it is because it is the user's will */
				level = BACKLIGHT_MIN;
			base->dimrequest = 1;
			break;
		case OP_DIM_OFF:
			base->savedlevel = base->currentlevel;
			level = BACKLIGHT_OFF;
			base->dimrequest = 2;
			break;
		case OP_DIM_RECOVER:
			level = base->savedlevel;
			if (level == BACKLIGHT_OFF)
				level = BACKLIGHT_MIN;    /* don't recover zero display brightness */
			base->dimrequest = 0;
			break;
		}

		base->targetlevel = level; /* brightness will be set in timer function */
		return level;
	}
	return -1;
}

void
display_framebuffer (int on)
{
	struct moddata_display *base = &modbase_display;
	int fd = -1;

	if (base->ctrl_fb) {
		if ((fd = open(base->fbdev, O_RDWR)) >= 0) {    /* open framebuffer device */
			ioctl(fd, FBIOBLANK, !on);
			close (fd);
		}
	}
}


