/* --------------------------------------------------------------------------
 * module_powersave.c
 * code for powermanagement module
 *
 * Copyright 2002 Matthias Grimm (joker@cymes.de).
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 * -------------------------------------------------------------------------*/

/* sleep reaction matrix:
 *                          disable     sleep
 *    sleep key pressed       x           x
 *    cover closed                        x
 *    after tlseep            x           x
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <string.h>
#include <initreq.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/kd.h>
#include "pbbinput.h"

#include <pbb.h>

#include "gettext_macros.h"
#include "input_manager.h"
#include "module_powersave.h"
#include "module_display.h"
#include "support.h"

/* --- private module data structure of module powermanagement --- */
struct moddata_power {
  struct modflags_power flags;
  struct powerprofile onAC;
  struct powerprofile onBattery;
  struct powerprofile *activeProfile;
  int powersource;		/* REDUNDANT copy from module pmac */
  int timeremaining;	/* REDUNDANT copy from module pmac */
  int mode;             /* module mode: awake, sleep, dim. */
  int sleeptime;         /* counter */
  int sleepcountdown;     /* another counter */
  int warnlevel;           /* current battery warnlevel */
  struct timeval tv;       /* time of key press */
  int sleepkeydelaytime;    /* delay in ms */
  unsigned short keysleep;  /* keycode for sleepfunction */
  unsigned short modsleep;  /* modifier for sleepfunction */
  int BWL1;
  int BWL2;
  int BWL3;
  int emergency;         /* emergency action at low battery condition */
  int cpuload_min;        /* cpuload minimum level to lock sleep */
  int cpuload_period;      /* period of time for that cpu activity mut be below cpuload */
  int netload_min;          /* netload minimum level to lock sleep */
  int netload_period;        /* period of time for that net activity mut be below netload */
  char *netload_dev;          /* device to observe for netload sleeplock */
  char *script_batcritical;    /* command to call if battery is critically low */
  char *script_profchanged;    /* command to call if profile has been changed */
  char *script_hdsetup;        /* command to call to tune the harddisk*/
} modbase_power;

/* --- interface functions of module powermanagement --- */

/* Init function called by the program init procedure. Function is known by
   the prginitab. It does module data initialisation and registers the module
   at the needed input queues. */

int
power_init (struct tagitem *taglist)
{
  struct moddata_power *base = &modbase_power;
  static char scriptbuffer1[STDBUFFERLEN];
  static char scriptbuffer2[STDBUFFERLEN];
  static char scriptbuffer3[STDBUFFERLEN];
  static char scriptbuffer4[STDBUFFERLEN];
  int rc;

  power_sync(); /* syncronise some data with module_pmac */

  scriptbuffer1[0]         = 0;             /* no default for script_batcritical */
  base->script_batcritical = scriptbuffer1;
  scriptbuffer2[0]         = 0;             /* no default for script_profchanged */
  base->script_profchanged = scriptbuffer2;
  scriptbuffer3[0]         = 0;             /* no default for script_hdsetup */
  base->script_hdsetup     = scriptbuffer3;
  base->mode               = MODE_AWAKE;

  /* default profile for running on AC power */
  base->onAC.flags.sleep      = base->flags.sleep_supported;
  base->onAC.flags.dim        = 1;
  base->onAC.flags.coversleep = base->flags.sleep_supported;
  base->onAC.Tsleep           = AC_TIME_SLEEP;
  base->onAC.Tdim             = AC_TIME_DIM;
  base->onAC.Thdoff           = AC_TIME_HDOFF;

  /* default profile for running on battery */
  base->onBattery.flags.sleep      = base->flags.sleep_supported;
  base->onBattery.flags.dim        = 1;
  base->onBattery.flags.coversleep = base->flags.sleep_supported;
  base->onBattery.Tsleep           = BATTERY_TIME_SLEEP;
  base->onBattery.Tdim             = BATTERY_TIME_DIM;
  base->onBattery.Thdoff           = BATTERY_TIME_HDOFF;

  /* definitions for immediatly sleep function */
  base->keysleep           = KEY_POWER;
  base->modsleep           = MOD_NONE;
  base->sleepkeydelaytime  = 0;         /* time delay till reaction */
  base->sleeptime          = 0;        /* counter for time till sleep */
  base->sleepcountdown     = 60;      /* counter for last 5 seconds before sleep */

  /* definitions for battery supervising functions */
  base->BWL1              = BAT_WARNLEVEL1;
  base->BWL2              = BAT_WARNLEVEL2;
  base->BWL3              = BAT_WARNLEVEL3;
  base->warnlevel         = 0;          /* current battery warnlevel */
  base->emergency         = base->flags.sleep_supported ? EMA_SLEEP : EMA_SIGNAL; /* default emergency action */
  base->flags.sigpwr_sent = 0;  /* currently no sigpwr in progress */

  /* definitions for sleeplocks */
  base->flags.cpuload_enable = 0;       /* disabled by default */
  base->flags.sleeplock_cpu  = 0;       /* lock open */
  base->cpuload_min          = 20;      /* lock closed if cpuload <20% */
  base->cpuload_period       = 60;      /*  for >60s, if enabled */
  base->flags.netload_enable = 0;       /* disabled by default */
  base->flags.sleeplock_net  = 0;       /* lock open */
  base->netload_min          = 4096;    /* lock closed if netload <4 KB/s */
  base->netload_period       = 60;      /*  for >60s, if enabled  */
  sprintf(scriptbuffer4, DEFAULT_NETLOAD_DEV);
  base->netload_dev          = scriptbuffer4;

  base->flags.heartbeat = 0;     /* heartbeat beep disabled */
  base->flags.heartbeat_enable = 1;  /* heartbeat beep is possible by default */ 

  if ((rc = power_handle_tags (MODE_CONFIG, taglist))) /* interprete configfile */
    return rc;         /* critcal error occured, abort program */

  ipc_protect_tag (TAG_SCRIPTBATCRITICAL);     /* tags only for privileged use */
  ipc_protect_tag (TAG_SCRIPTPROFCHANGED);
  ipc_protect_tag (TAG_SCRIPTHDSETUP);

  register_function (KBDQUEUE, power_keyboard);
  register_function (MOUSEQUEUE, power_awake);
  register_function (T100QUEUE, power_timer);
  register_function (T1000QUEUE, power_timer1000);
  register_function (QUERYQUEUE, power_query);
  register_function (CONFIGQUEUE, power_configure);
  return 0;
}

int
power_exit ()
{
	return 0;
}

void
power_keyboard (struct tagitem *taglist)
{
	struct moddata_power *base = &modbase_power;
	int code, value, mod;

	code = (int) tagfind (taglist, TAG_KEYCODE, 0);
	value = (int) tagfind (taglist, TAG_KEYREPEAT, 0);
	mod = (int) tagfind (taglist, TAG_MODIFIER, 0);

	if (value)
		if (value == 1) {
			power_awake (taglist);
			base->flags.sleeptriggered = 0;
		}
		if ((code == base->keysleep) && (mod == base->modsleep))
			if ((base->sleepkeydelaytime == 0) || (keydelayms(&base->tv, value, base->sleepkeydelaytime)))
				if (base->flags.sleeptriggered == 0) {
					base->flags.sleeptriggered = 1;     /* sleep should be triggered only once */
					singletag_to_clients(WARNING, TAG_SLEEPINSECONDS, 0);
					sleep(1);   /* give clients a chance to react */
					power_trigger_sleep (base->activeProfile->flags.sleep);
				}
}

/* This function will be called on every keyboard or mouse action. With
   other words the user is active if this function is called so sleeptimer
   is reset here. */

void
power_awake (struct tagitem *taglist)
{
	struct moddata_power *base = &modbase_power;

	if (base->mode != MODE_AWAKE) {
		process_queue_single (CONFIGQUEUE, TAG_BRIGHTNESSOP, OP_DIM_RECOVER);
		base->flags.heartbeat = 0; /* disable heartbeat beep */
		base->mode = MODE_AWAKE;
	}
	base->sleeptime = 0;
}

void
power_sync ()
{
	struct moddata_power *base = &modbase_power;
	struct tagitem args[4];

	taglist_init (args);
	taglist_add (args, TAG_SLEEPSUPPORTED, 0);
	taglist_add (args, TAG_POWERSOURCE, 0);
	taglist_add (args, TAG_TIMEREMAINING, -1);
	process_queue (QUERYQUEUE, args);
	base->flags.sleep_supported = tagfind (args, TAG_SLEEPSUPPORTED, 0);
	base->powersource = tagfind (args, TAG_POWERSOURCE, 0);
	base->timeremaining = tagfind (args, TAG_TIMEREMAINING, -1);
	base->activeProfile = base->powersource ? &base->onAC : &base->onBattery;
}

void
power_query (struct tagitem *taglist)
{
  power_handle_tags (MODE_QUERY, taglist);
}

void
power_configure (struct tagitem *taglist)
{
  power_handle_tags (MODE_CONFIG, taglist);
}

int
power_handle_tags (int cfgure, struct tagitem *taglist)
{
	struct moddata_power *base = &modbase_power;
	int err, rc = 0;

	while (taglist->tag != TAG_END) {
		switch (taglist->tag) {
		case TAG_SCRIPTBATCRITICAL:
			if (cfgure) {
				if ((check_script ((char *) taglist->data, "")) == 0) {
					if ((err = copy_path ((char *) taglist->data, base->script_batcritical, TYPE_FILE, CPFLG_PBBONLY)))
						rc = tagerror (taglist, err);
				} else {
					print_error(_("ERROR: Too many formatsigns. None allowed in TAG_SCRIPTBATCRITICAL.\n"));
					rc = tagerror (taglist, E_FORMAT);
				}
			} else
				taglist->data = (long) base->script_batcritical;
			break;
		case TAG_SCRIPTPROFCHANGED:
			if (cfgure) {
				if ((check_script ((char *) taglist->data, "ss")) == 0) {
					if ((err = copy_path ((char *) taglist->data, base->script_profchanged, TYPE_FILE, CPFLG_PBBONLY)))
						rc = tagerror (taglist, err);
				} else {
					print_error(_("ERROR: Too many formatsigns. Max two %%s allowed in TAG_SCRIPTPROFCHANGED.\n"));
					rc = tagerror (taglist, E_FORMAT);
				}
			} else
				taglist->data = (long) base->script_profchanged;
			break;
		case TAG_SCRIPTHDSETUP:
			if (cfgure) {
				if ((check_script ((char *) taglist->data, "d")) == 0) {
					if ((err = copy_path ((char *) taglist->data, base->script_hdsetup, TYPE_FILE, CPFLG_PBBONLY)))
						tagerror (taglist, err);
				} else {
					print_error(_("ERROR: Too many formatsigns. Only one %%d allowed in TAG_SCRIPTHDSETUP.\n"));
					tagerror (taglist, E_FORMAT);
				}
			} else
				taglist->data = (long) base->script_hdsetup;
			break;
		case TAG_ONACSLEEP:
			if (cfgure) {
				if (base->flags.sleep_supported)
					base->onAC.flags.sleep = taglist->data;
			} else
				taglist->data = base->onAC.flags.sleep;
			break;
		case TAG_ONACDIM:
			if (cfgure)	base->onAC.flags.dim = taglist->data;
			else		taglist->data = base->onAC.flags.dim;
			break;
		case TAG_ONACCOVERSLEEP:
			if (cfgure)	base->onAC.flags.coversleep = taglist->data;
			else		taglist->data = base->onAC.flags.coversleep;
			break;
		case TAG_ONACTIMESLEEP:
			if (cfgure)	base->onAC.Tsleep = taglist->data;
			else		taglist->data = base->onAC.Tsleep;
			break;
		case TAG_ONACTIMEDIM:
			if (cfgure)	base->onAC.Tdim = taglist->data;
			else		taglist->data = base->onAC.Tdim;
			break;
		case TAG_ONACTIMEHDOFF:
			if (cfgure) {
				base->onAC.Thdoff = taglist->data;
				if (base->activeProfile == &base->onAC && strlen(base->script_hdsetup) > 0)
					call_script (base->script_hdsetup, base->activeProfile->Thdoff);
			} else
				taglist->data = base->onAC.Thdoff;
			break;
		case TAG_ONBATTERYSLEEP:
			if (cfgure) {
				if (base->flags.sleep_supported)
					base->onBattery.flags.sleep = taglist->data;
			} else
				taglist->data = base->onBattery.flags.sleep;
			break;
		case TAG_ONBATTERYDIM:
			if (cfgure)	base->onBattery.flags.dim = taglist->data;
			else		taglist->data = base->onBattery.flags.dim;
			break;
		case TAG_ONBATTERYCOVERSLEEP:
			if (cfgure)	base->onBattery.flags.coversleep = taglist->data;
			else		taglist->data = base->onBattery.flags.coversleep;
			break;
		case TAG_ONBATTERYTIMESLEEP:
			if (cfgure)	base->onBattery.Tsleep = taglist->data;
			else		taglist->data = base->onBattery.Tsleep;
			break;
		case TAG_ONBATTERYTIMEDIM:
			if (cfgure)	base->onBattery.Tdim = taglist->data;
			else		taglist->data = base->onBattery.Tdim;
			break;
		case TAG_ONBATTERYTIMEHDOFF:
			if (cfgure) {
				base->onBattery.Thdoff = taglist->data;
				if (base->activeProfile == &base->onBattery && strlen(base->script_hdsetup) > 0)
					call_script (base->script_hdsetup, base->activeProfile->Thdoff);
			} else
				taglist->data = base->onBattery.Thdoff;
			break;
		case TAG_SLEEPKEY:
			if (cfgure)	base->keysleep = taglist->data;
			else		taglist->data = base->keysleep;
			break;
		case TAG_SLEEPMOD:
			if (cfgure)	base->modsleep = taglist->data;
			else		taglist->data = base->modsleep;
			break;
		case TAG_SLEEPKEYDELAY:
			if (cfgure)	base->sleepkeydelaytime = taglist->data;
			else		taglist->data = base->sleepkeydelaytime;
			break;
		case TAG_BWLFIRST:
			if (cfgure)	base->BWL1 = taglist->data;
			else		taglist->data = base->BWL1;
			break;
		case TAG_BWLSECOND:
			if (cfgure)	base->BWL2 = taglist->data;
			else		taglist->data = base->BWL2;
			break;
		case TAG_BWLLAST:
			if (cfgure)	base->BWL3 = taglist->data;
			else		taglist->data = base->BWL3;
			break;
		case TAG_CURRENTBWL:
			if (cfgure)	tagerror (taglist, E_NOWRITE);
			else		taglist->data = base->warnlevel;
			break;
		case TAG_CPULOADSLEEPLOCK:
			if (cfgure) {
				base->flags.cpuload_enable = taglist->data;
				base->flags.sleeplock_cpu = taglist->data;
			} else
				taglist->data = base->flags.cpuload_enable;
			break;
		case TAG_CPULOADMIN:
			if (cfgure)	base->cpuload_min = taglist->data;
			else		taglist->data = base->cpuload_min;
			break;
		case TAG_CPULOADPERIOD:
			if (cfgure)	base->cpuload_period = taglist->data;
			else		taglist->data = base->cpuload_period;
			break;
		case TAG_NETLOADSLEEPLOCK:
			if (cfgure) {
				base->flags.netload_enable = taglist->data;
				base->flags.sleeplock_net = taglist->data;
			} else
				taglist->data = base->flags.netload_enable;
			break;
		case TAG_NETLOADMIN:
			if (cfgure)	base->netload_min = taglist->data;
			else		taglist->data = base->netload_min;
			break;
		case TAG_NETLOADPERIOD:
			if (cfgure)	base->netload_period = taglist->data;
			else		taglist->data = base->netload_period;
			break;
		case TAG_NETLOADDEV:
			if (cfgure)	strncpy (base->netload_dev, (char *) taglist->data, STDBUFFERLEN-1);
			else		taglist->data = (long) base->netload_dev;
			break;
		case TAG_EMERGENCYACTION:
			if (cfgure) {
				if (taglist->data == EMA_SLEEP) {
					if (base->flags.sleep_supported) /* sleep supported? */
						base->emergency = EMA_SLEEP; /* then set to EMA_SLEEP */
					else
						tagerror (taglist, E_NOSUPPORT); /* otherwise send error */
				} else
					base->emergency = taglist->data;
			} else
				taglist->data = (long) base->emergency;
			break;
		case TAG_HEARTBEATBEEP:
			if (cfgure)	base->flags.heartbeat_enable = taglist->data;
			else		taglist->data = base->flags.heartbeat_enable;
			break;
		case TAG_POWERCHANGED:  /* private tag */
			if (cfgure) {
				power_awake (NULL);
				base->powersource = taglist->data;
				if (base->powersource) {
					base->activeProfile = &base->onAC;
					call_script (base->script_profchanged, "profile", "ac");
				} else {
					base->activeProfile = &base->onBattery;
					call_script (base->script_profchanged, "profile", "battery");
				}
				call_script (base->script_hdsetup, base->activeProfile->Thdoff);
				power_check_battery (base->timeremaining, base->powersource);
			}
			break;
		case TAG_TIMECHANGED:  /* private tag */
			if (cfgure) {
				base->timeremaining = taglist->data;
				power_check_battery (base->timeremaining, base->powersource);
			}
			break;
		case TAG_PREPAREFORSLEEP:  /* private tag */
			if (cfgure)	call_script (base->script_profchanged, "sleep", base->powersource ? "ac" : "battery");
			break;
		case TAG_WAKEUPFROMSLEEP:  /* private tag */
			if (cfgure) {
				power_awake (NULL);
				power_sync ();  /* syncronise redundant data from module_pmac */
				call_script (base->script_profchanged, "wakeup", base->powersource ? "ac" : "battery");
				power_check_battery (-1, base->powersource);
			}
			break;
		case TAG_COVERSTATUS:  /* private tag */
			if (cfgure) {
				if (taglist->data)
					power_awake (NULL);
				else
					power_trigger_sleep (base->activeProfile->flags.coversleep);
			}
			break;
		case TAG_GOTOSLEEP:
			if (cfgure)	power_trigger_sleep (base->activeProfile->flags.sleep);
			else		tagerror (taglist, E_NOREAD);
			break;
		}
		taglist++;
	}
	return rc;
}

/* This routine will be called every 10ms */

void
power_timer (struct tagitem *taglist)
{
  struct moddata_power *base = &modbase_power;
  int n;

  base->sleeptime++;
  switch (base->mode) {
  case MODE_AWAKE:
    if ((base->sleeptime > base->activeProfile->Tdim)
        && base->activeProfile->flags.dim) { /* time to dim the display? */
      process_queue_single (CONFIGQUEUE, TAG_BRIGHTNESSOP, OP_DIM_LIGHT); /* then dim */
      base->mode = MODE_DIM;
    }
  case MODE_DIM:
    if ((base->sleeptime > base->activeProfile->Tsleep)
        && !base->flags.sleeplock_cpu
        && !base->flags.sleeplock_net) { /* time to sleep? */
      base->mode = MODE_DOZY;
      base->sleepcountdown = 60;
    }
    break;
  case MODE_DOZY:
    n = --base->sleepcountdown / 10;       /* time left in seconds */
    if (n == 0)
      power_trigger_sleep (base->activeProfile->flags.sleep);
    else
      singletag_to_clients(WARNING, TAG_SLEEPINSECONDS, n);
  case MODE_SLEEP:
    base->sleeptime--;
    break;
  }
}

void
power_timer1000 (struct tagitem *taglist)
{
	struct moddata_power *base = &modbase_power;
	static int cputimer = 0, nettimer = 0, beeptimer = 0;
	int cpuload = 0, netload = 0;

	if (base->flags.cpuload_enable) {
		cpuload = power_read_cpu ();
		if (cpuload < base->cpuload_min) {
			if (base->flags.sleeplock_cpu) {   /* skip following if lock is already open */
				cputimer++;
				if (cputimer > base->cpuload_period)
					base->flags.sleeplock_cpu = 0;   /* open lock */
			}
		} else {
			cputimer = 0;
			base->flags.sleeplock_cpu = 1;   /* close lock */
		}
	} else
		cputimer = 0;

	if (base->flags.netload_enable) {
		netload = power_read_net (base->netload_dev);
		if (netload < base->netload_min) {
			if (base->flags.sleeplock_net) {   /* skip following if lock is already open */
				nettimer++;
				if (nettimer > base->netload_period)
					base->flags.sleeplock_net = 0;   /* open lock */
			}
		} else {
			nettimer = 0;
			base->flags.sleeplock_net = 1;   /* close lock */
		}
	} else
		nettimer = 0;

	if (base->flags.heartbeat_enable) {  /* user wants heartbeat beep ? */
		if (base->flags.heartbeat) {
			beeptimer++;
			if (beeptimer > BEEPINTERVAL) {
				power_beep();
				beeptimer = 0;
			}
		} else
			beeptimer = 0;
	} else
		beeptimer = 0;

#if 0
#ifdef DEBUG
	printf ("cpu: ");
	if (base->flags.cpuload_enable)
		printf ("%2d%% (lock %s)", cpuload, base->flags.sleeplock_cpu ? "closed" : "open  ");
	else
		printf ("disabled");

	printf (",  net: ");
	if (base->flags.netload_enable)
		printf ("%8d B/s (lock %s)", netload, base->flags.sleeplock_net ? "closed" : "open  ");
	else
		printf ("disabled");
	printf ("\r");
	fflush(stdout);
#endif
#endif
}

void
power_check_battery (int time, int psource)
{
	struct moddata_power *base = &modbase_power;
	struct tagitem taglist[3];

	int warnlevel[] = {1200, base->BWL1 + BAT_HYSTERESIS,
	                       base->BWL1, base->BWL2 + BAT_HYSTERESIS,
	                       base->BWL2, base->BWL3 + BAT_HYSTERESIS,
	                       base->BWL3, 1,
	                       1,          0};
	int n, wl;

	/* if pbbuttons was started with AC plugged in, timeleft would
	 * remain at -1 and after AC pulled out the value wasn't changed
	 * quick enough so that the warning level 4 was set due to
	 * -1 / 60 = 0 -> wl = 4 -> shutdown. To prevent this the timeleft
	 * value of -1 is filtered out here
	 */

	if (psource || time == -1) { /* if battery is charging or time is invalid */
		wl = 0;                   /* we don't need to calculate warnlevels */
		if (base->warnlevel) {      /* but maybe have to inform clients */
			taglist_init (taglist);
			taglist_add (taglist, TAG_CURRENTBWL, wl);
			distribute_to_clients(WARNING, taglist);
		}
	} else {
		time /= 60;           /* calculate minutes */
		wl = base->warnlevel;
		if (time >= 0) {
			for (n=4; n >= 0; n--)
				if (time < warnlevel[2*n])
					if (time >= warnlevel[2*n+1]) {
						wl = n;
						break;
					}

			if (wl > base->warnlevel) {  /* warnlevel increased ? */
				taglist_init (taglist);
				taglist_add (taglist, TAG_CURRENTBWL, wl);
				taglist_add (taglist, TAG_TIMEREMAINING, time);
				distribute_to_clients(WARNING, taglist);
				if (wl == 4) {             /* oh oh, battery is realy low */
					switch (base->emergency) {
					case EMA_SLEEP:
						process_queue_single (CONFIGQUEUE, TAG_REQUESTSLEEP, 0);
						break;
					case EMA_SIGNAL:
						if ((send_sigpwr('F')) != 0) { /* send init a SIGPWR signal */
							if (base->flags.sleep_supported) { /* fallback to sleep if possible */
								base->emergency = EMA_SLEEP;
								process_queue_single (CONFIGQUEUE, TAG_REQUESTSLEEP, 0);
							} else {              /* otherwise use the last way out */
								base->emergency = EMA_COMMAND;
								call_script (base->script_batcritical);
								sleep (10);
							}
						} else {
							base->flags.sigpwr_sent = 1;  /* sigpwr in progress */
							sleep(5);  /* give INIT a chance to react */
						}
						break;
					case EMA_COMMAND:
						call_script (base->script_batcritical); /* usually shutdown */
						sleep(10);   /* give the script a chance to work */
						break;
					}
				}
			} else if (wl < base->warnlevel && base->flags.sigpwr_sent) { /* battery has recovered */
				send_sigpwr('O');             /* abort sigpwr action */
				base->flags.sigpwr_sent = 0; /* sigpwr no longer in progress */
			}
		}
	}
	base->warnlevel = wl;
}

/* This function requests sleep mode and take user configurations
    like enable_blank into account. */

void
power_trigger_sleep (int flag)
{
	struct moddata_power *base = &modbase_power;

	base->mode = MODE_SLEEP;
	if (flag && base->flags.sleep_supported)
		process_queue_single (CONFIGQUEUE, TAG_REQUESTSLEEP, 0);
	else {
		process_queue_single (CONFIGQUEUE,TAG_BRIGHTNESSOP, OP_DIM_OFF);
		if (base->flags.heartbeat_enable) {  /* user wants heartbeat beep ? */
			base->flags.heartbeat = 1; /* enable heartbeat beep */
			power_beep();
		}
	}
}

int
check_script (char *name, char *allowed)
{
  while (*name != 0) {
    if (*name++ == '%')
      if ((*allowed == 0) || (*name++ != *allowed++))
        return -1;
  }
  return 0;
}

int
send_sigpwr (char mode)
{
  struct init_request initreq;
  FILE *fd;
  int err = -1;

  initreq.magic     = INIT_MAGIC;
  initreq.cmd       = mode == 'F' ? INIT_CMD_POWERFAIL : INIT_CMD_POWEROK;
  initreq.sleeptime = 5;

  if ((fd = fopen(INIT_FIFO, "r+")) != NULL) {
    err = 0;
    if ((fwrite(&initreq, sizeof(initreq), 1, fd)) != 1)
      err = -1;
    fclose(fd);
  }
  if (err == -1) {
    if ((fd = fopen("/etc/powerstatus", "w")) != NULL) {
      err = 0;
      if ((fwrite(&mode, 1, 1, fd)) != 1)
        err = -1;
      fclose(fd);
    }
    err = kill(1, SIGPWR);  /* send SIGPWR to the INIT process */
  }
  return err;
}

/* This function reads /proc/stat and calculates the current cpuload.
    If an error occured, the return value would be '0', so that no program
    features espacially sleep will be blocked. */

int
power_read_cpu (void)
{
	char buffer[1024], *token;
	static ulong pvalue = 0, ptotal;
	ulong value = 0, total = 1, user, nice, sys, idle;
	FILE *fd;

	if ((fd = fopen ("/proc/stat","r"))) {
		while (fgets (buffer, sizeof (buffer), fd))
			if ((token = strtok (buffer," \t\n"))) {
				if (!strncmp ("cpu", token, 3)) {
					sscanf(strtok (0, "\n"), "%lu %lu %lu %lu",
					         &user, &nice, &sys, &idle);
					break;
				} else
					strtok (0,"\n");
			}
		fclose(fd);

		if (pvalue) {
			value = user + sys + nice - pvalue;
			total = user + sys + nice + idle - ptotal;
		}
		pvalue = user + sys + nice;
		ptotal = pvalue + idle;
		return (int) (value * 100 / total);
	}
	return 0;
}

/* This function reads /proc/net/dev and calculates the current trafic to and
    from net. Only the given network device will be observed and incomming
    and outgoing trafic would be combined to one single bytes/second value.
    If an error occured, the return value would be '0', so that no program
    features espacially sleep will be blocked. */

int
power_read_net (char *dev)
{
	static ulong ptrafic = 0;
	char buffer[1024], *token;
	ulong dummy, bytesread = 0, byteswrite = 0, trafic = 0;
	FILE *fd;

	if ((fd = fopen ("/proc/net/dev","r"))) {
		while (fgets (buffer, sizeof (buffer), fd))
			if ((token = strtok (buffer,":\n"))) {
				cleanup_buffer (token);
				if (!strncmp (dev, token, strlen (dev))) {
					sscanf(strtok (0, "\n"), "%lu %lu %lu %lu %lu %lu %lu %lu %lu",
					         &bytesread, &dummy, &dummy, &dummy, &dummy,
						 &dummy, &dummy, &dummy, &byteswrite);
					break;
				} else
					strtok (0,"\n");
			}
		fclose(fd);

		if (ptrafic)
			trafic = bytesread + byteswrite - ptrafic;
		ptrafic = bytesread + byteswrite;
		return (int) trafic;
	}
	return 0;
}

void
power_beep()
{
	int arg, fd;
	unsigned int ms = 150;
	unsigned int freq = 5000;

	if ((fd = open("/dev/console", O_RDWR)) >= 0) {
		arg = (ms << 16) | freq;
		ioctl(fd, KDMKTONE, arg);
		usleep(ms*1000);
		close (fd);
	}
}
